"""
Date helpers.
"""
from datetime import date


def is_weekend(day: date):
    """
    Say if this day is a weekend or not.

    Parameters
    ----------
        day: date
            Day.

    Returns
    -------
        bool
            True if is weekend.
    """
    return day.weekday() in [5, 6]



"""
Backtester
"""

from datetime import timedelta
import os
from pprint import pprint
import uuid

import numpy as np
import pandas as pd
from tqdm import tqdm

from .broker import Broker
from .contract import Contract
from .market_data import MarketData
from ..utils.dates import is_weekend


TWELVE_MONTHS = 250


from dataclasses import dataclass


# pylint: disable=too-many-instance-attributes
@dataclass
class BacktesterParameters:
    """
    Backtester parameters.
    """

    def __init__(self):
        self.cash = 0
        self.custom = {}
        self.end_date = None
        self.leverage = 1
        self.live = False
        self.no_check = False
        self.plot = True
        self.suffix = ""
        self.start_date = None
        self.tickers = []


# pylint:
class Backtester:
    """
    This class implements the backtester.
    The function run() iterates on every day of the backtesting period and
    calls next() and next_iterators() where respectively the decisions and
    indicators computations happen.
    """

    def __init__(
        self,
        params,
    ):
        self.broker = Broker(params.cash, params.live, params.no_check)
        self.market_data = MarketData()
        self.data = []
        self.dates = []
        self.day = None
        self._params = params
        self.nav = params.cash

    def compute_kelly(self):
        """
        Compute the Kelly ratio, an indicator for strategy max leverage.

        Parameters
        ----------

        Returns
        -------
            float
                Kelly ratio.
        """
        dfm = pd.DataFrame(data=self.data, index=self.dates)
        returns = np.diff(np.log(dfm.Nav))
        kelly = np.nanmean(returns) / np.power(np.nanstd(returns), 2)
        return kelly

    def compute_mean(self):
        """
        Compute the mean of daily returns, annualized.

        Parameters
        ----------

        Returns
        -------
            float
                Mean of daily returns, annualized.
        """
        dfm = pd.DataFrame(data=self.data, index=self.dates)
        returns = np.diff(np.log(dfm.Nav))
        mean = np.nanmean(returns) * TWELVE_MONTHS
        return mean

    def _plot_nav(self):
        dfm = pd.DataFrame(data=self.data, index=self.dates)
        fname = ",".join(self._params.tickers)
        if len(fname) > 30:
            fname = str(uuid.uuid3(uuid.NAMESPACE_URL, fname))
        filename = (
            fname
            + f".{self._params.leverage}.{self._params.end_date.isoformat()}"
            + f".{self._params.suffix}.png"
        )
        path = os.path.join(os.getenv("HOME"), "Downloads", filename)
        dfm[["Nav"]].plot(logy=True).get_figure().savefig(path)

    def compute_sharpe_ratio(self):
        """
        Compute the Sharpe ratio.

        Parameters
        ----------

        Returns
        -------
            float
                Sharpe ratio.
        """
        dfm = pd.DataFrame(data=self.data, index=self.dates)
        returns = np.diff(np.log(dfm.Nav))
        sharpe_ratio = np.nanmean(returns) / np.nanstd(returns) * np.sqrt(250)
        return sharpe_ratio

    def compute_std(self):
        """
        Compute the standard deviation of daily returns, annualized (ie. the volatility).

        Parameters
        ----------

        Returns
        -------
            float
                Standard deviation of daily returns, annualized.
        """
        dfm = pd.DataFrame(data=self.data, index=self.dates)
        returns = np.diff(np.log(dfm.Nav))
        std = np.nanstd(returns) * np.sqrt(TWELVE_MONTHS)
        return std

    def _has_not_enough_active_contracts(self):
        for ticker in self._params.tickers:
            active_contracts = Contract(ticker=ticker, day=self.day).chain
            if active_contracts.shape[0] < 2:
                return ticker
        return None

    def run(self):
        """
        This function iterate on the backtest period and call
        the next function where the logic of the strategy is implemented.

        Parameters
        ----------

        Returns
        -------
        """
        delta = self._params.end_date - self._params.start_date
        for i in tqdm(range(delta.days + 1)):
            self.day = self._params.start_date + timedelta(days=i)
            not_enough_active_contracts = self._has_not_enough_active_contracts()
            if not_enough_active_contracts is not None:
                raise Exception(
                    f"Update future-expiry/{not_enough_active_contracts}.csv in Minio"
                )
            if is_weekend(self.day):
                continue
            self.broker.next(self.day)
            self.next()
            self.next_indicators()
            self.dates.append(self.day)
            nav = self.broker.nav
            if not np.isnan(nav):
                self.nav = nav
            data = {}
            data["Nav"] = nav
            self.data.append(data)
        if not self._params.live and self._params.plot:
            self._plot_nav()
        kelly = self.compute_kelly()
        mean = self.compute_mean()
        sharpe = self.compute_sharpe_ratio()
        std = self.compute_std()
        pprint(
            {
                "day": self.day,
                "mean": mean,
                "std": std,
                "kelly": kelly,
                "sharpe": sharpe,
            }
        )

    def next(self):
        """
        This function contains the logic of the strategy: decision, update of positions, etc.
        It is called on every new day.

        Parameters
        ----------

        Returns
        -------
        """
        raise Exception("To be implemented in the child class")

    def next_indicators(self):
        """
        This function contains the logic of the calculation of the indicators.
        It is called on every new day, after next().

        Parameters
        ----------

        Returns
        -------
        """
        raise Exception("To be implemented in the child class")



"""
Client for the API https://marketdata.edarchimbaud.com
"""
from datetime import date
import os

import pandas as pd
import requests

from .constants import LAST_MODIFIED


class Client:
    """
    Client for the API https://marketdata.edarchimbaud.com
    """

    def __init__(self):
        self.api_url = "https://" + os.getenv("DATA_DOMAIN")
        self.headers = {"Authorization": os.getenv("DATA_SECRET_KEY")}

    def patch_bucket(self, bucket_name: str, action: str = LAST_MODIFIED):
        """
        Send a patch query to modify a bucket.

        Parameters
        ----------
            bucket_name: str
                Bucket name.

            action: str
                Action name, either empty or last-modified.

        Returns
        -------
        """
        requests.patch(
            f"{self.api_url}/private/bucket",
            headers=self.headers,
            params={"bucket_name": bucket_name, "action": action},
        )

    def get_daily_borrowing_rates(self, day: date):
        """
        Get daily borrowing rates.

        Parameters
        ----------
            day: date
                Day of the rates to query.

        Returns
        -------
            DataFrame
                Borrowing rates.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/daily/borrowing-rates",
            headers=self.headers,
            params={
                "day": day.isoformat(),
            },
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        dfm = dfm.set_index(["Date", "RIC"])
        return dfm, error

    def get_daily_factor(
        self, name: str, ticker: str, start_date: date, end_date: date
    ):
        """
        Get daily factor.

        Parameters
        ----------
            name: str
                Name of the factor to get. Either:
                    carry/bond
                    carry/commodity
                    carry/currency
                    carry/equity
                    cot
                    currency
                    nav/long
                    nav/short
                    news/headlines
                    news/stories
                    roll-return
                    splits

            ticker: str
                Instrument ticker.

            start_date: date
                Start date.

            end_date: date
                End date.

        Returns
        -------
            DataFrame
                Factor data.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/daily/factor/{name}",
            headers=self.headers,
            params={
                "ticker": ticker,
                "start_date": start_date.isoformat(),
                "end_date": end_date.isoformat(),
            },
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        dfm = dfm.set_index(["Date", "Stem"])
        return dfm, error

    def get_daily_ohlcv(self, ric: str, start_date: date, end_date: date):
        """
        Get daily Open High Low Close Volume.

        Parameters
        ----------
            ric: str
                Instrument RIC.

            start_date: date
                Start date.

            end_date: date
                End date.

        Returns
        -------
            DataFrame
                Open High Low Close Volume data.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/daily/ohlcv",
            headers=self.headers,
            params={
                "ric": ric,
                "start_date": start_date.isoformat(),
                "end_date": end_date.isoformat(),
            },
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        dfm = dfm.set_index(["Date", "RIC"])
        return dfm, error

    def get_daily_risk_free_rate(self, ric: str, start_date: date, end_date: date):
        """
        Get daily risk free rate.

        Parameters
        ----------
            ric: str
                Instrument RIC.

            start_date: date
                Start date.

            end_date: date
                End date.

        Returns
        -------
            DataFrame
                Risk free rate data.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/daily/risk-free-rate",
            headers=self.headers,
            params={
                "ric": ric,
                "start_date": start_date.isoformat(),
                "end_date": end_date.isoformat(),
            },
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        dfm = dfm.set_index(["Date", "RIC"])
        return dfm, error

    def get_dataset(self, ticker: str, start_date: date, end_date: date):
        """
        Get dataset of market features and targets.

        Parameters
        ----------
            ticker: str
                Instrument ticker.

            start_date: date
                Start date.

            end_date: date
                End date.

        Returns
        -------
            DataFrame
                Dataset data.

            str
                Error message.
        """
        is_authenticated = bool(os.getenv("DATA_SECRET_KEY"))
        url = (
            f"{self.api_url}/private/dataset"
            if is_authenticated
            else f"{self.api_url}/public/dataset"
        )
        response = requests.get(
            url,
            headers=self.headers,
            params={
                "ticker": ticker,
                "start_date": start_date.isoformat(),
                "end_date": end_date.isoformat(),
            },
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        return dfm, error

    def get_expiry_calendar(self, ticker: str, download: bool = False):
        """
        Get expiry calendar.

        Parameters
        ----------
            ticker: str
                Instrument ticker.

            download: bool
                For download if True. Otherwise uses cache if possible.

        Returns
        -------
            DataFrame
                Expiry calendar data.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/expiry-calendar",
            headers=self.headers,
            params={"ticker": ticker, "download": download},
        )
        response_json = response.json()
        error = response_json["error"]
        data = response_json["data"]
        if data is None:
            return None, error
        dfm = pd.DataFrame.from_dict(data)
        return dfm, error

    def get_health_ric(self, ric: str):
        """
        Check if RIC is still live or not.

        Parameters
        ----------
            ric: str
                Instrument RIC.

        Returns
        -------
            bool
                Either True if the RIC is live or False if not.

            str
                Error message.
        """
        response = requests.get(
            f"{self.api_url}/private/health/ric",
            headers=self.headers,
            params={
                "ric": ric,
            },
        )
        response_json = response.json()
        data = response_json["data"]
        error = response_json["error"]
        return data, error

    def get_tickers(self):
        """
        Get the list of instruments of the perimeter and their parameters.

        Parameters
        ----------

        Returns
        -------
            object
                List of instruments and their parameters.

            str
                Error message.
        """
        response = requests.get(f"{self.api_url}/private/tickers", headers=self.headers)
        response_json = response.json()
        data = response_json.get("data", [])
        error = response_json["error"]
        return data, error



"""
Margin module.
"""
from datetime import date
import numpy as np

from .contract import Contract
from .forex import Forex
from .market_data import MarketData
from ..data.constants import FUTURES


class Margin:
    """
    Margin.
    """

    def __init__(self):
        self.cache = {}
        self.forex = Forex()
        self.market_data = MarketData()

    def __adjustment_factor(self, ticker: str, day: date):
        # update of adjustment factor is done yearly
        key = f"{ticker} {day.year}"
        if key in self.cache:
            return self.cache[key]
        contract, _ = Contract(ticker=ticker, day=day).front_contract
        if not self.market_data.is_trading_day(contract=contract, day=day):
            return np.NaN
        row = self.market_data.bardata(contract=contract, day=day)
        ref_date = self._get_ref_date(ticker=ticker)
        ref_contract, _ = Contract(ticker=ticker, day=ref_date).front_contract
        row_ref = self.market_data.bardata(contract=ref_contract, day=ref_date)
        self.cache[key] = row["Close"][0] / row_ref["Close"][0]
        return self.cache[key]

    def _get_ref_date(self, ticker: str):
        if ticker in ["HTE", "MBT"]:
            return date(2021, 8, 18)
        default_ref_date = date(2020, 1, 6)
        return default_ref_date

    def overnight_initial_future(self, ticker: str, day: date):
        """
        Get the initial overnight margin limit.

        Parameters
        ----------
            ticker: str
                Ticker of the instrument.

            day: date
                Date to get the initial overnight margin limit.

        Returns
        -------
            float
                Initial overnight margin limit.
        """

        currency = FUTURES[ticker]["Currency"]
        return (
            FUTURES[ticker]["OvernightInitial"]
            * self.__adjustment_factor(ticker=ticker, day=day)
            * self.forex.to_usd(currency, day)
        )

    def overnight_maintenance_future(self, ticker: str, day: date):
        """
        Get the maintenance overnight margin limit.

        Parameters
        ----------
            ticker: str
                Ticker of the instrument.

            day: date
                Date to get the maintenance overnight margin limit.

        Returns
        -------
            float
                Maintenance overnight margin limit.
        """
        currency = FUTURES[ticker]["Currency"]
        return (
            FUTURES[ticker]["OvernightMaintenance"]
            * self.__adjustment_factor(ticker=ticker, day=day)
            * self.forex.to_usd(currency, day)
        )



"""
Market data access.
"""
from datetime import date, timedelta

from methodtools import lru_cache
import numpy as np
from pandas_market_calendars import get_calendar

from .contract import Contract
from ..data.client import Client
from ..data.constants import FUTURES
from ..utils.dates import is_weekend


LIBOR_BEFORE_2001 = 6.65125
MAXIMUM_NUMBER_OF_DAYS_BEFORE_EXPIRY = 40


class MarketData:
    """
    Market data access.
    """

    def __init__(self):
        pass

    def bardata(self, contract: Contract, day: date):
        """
        Get bar data.

        Parameters
        ----------
            ric: str
                RIC of the instrument.

            day: date
                Day collect data for.

        Returns
        -------
            DataFrame
                Collected bar data.
        """
        data, err = self.get_future_ohlcv_for_day(contract=contract, day=day)
        if err:
            raise Exception(err["message"])
        data = data.fillna(value=np.nan)
        if (
            np.isnan(data.Close[0])
            and not np.isnan(data.Volume[0])
            and not np.all(np.isnan(data[["Open", "High", "Low"]].values[0]))
        ):
            data.Close = np.nanmedian(data[["Open", "High", "Low"]].values[0])
        return data

    @lru_cache()
    @staticmethod
    def __get_future_ohlcv(ric, start_date, end_date):
        """
        Get OHLCV data for a future.

        Parameters
        ----------
            ric: str
                RIC of the instrument.

            start_date: date
                Start date of the period to collect data from.

            end_date: date
                End date of the period to collect data from.

        Returns
        -------
            DataFrame
                Future data.
            str
                Error message.
        """
        dfm, error = Client().get_daily_ohlcv(ric, start_date, end_date)
        if dfm is None:
            return None, error
        dfm.reset_index(drop=False, inplace=True)
        dfm.Date = dfm.Date.apply(lambda x: x[:10])
        dfm.drop(columns=["RIC"], inplace=True)
        dfm.set_index("Date", drop=True, inplace=True)
        return dfm, error

    def get_future_ohlcv_for_day(self, contract: Contract, day: date):
        """
        Get OHLCV data for a future and for a specific day.

        Parameters
        ----------
            contract: Contract
                Contract of the instrument.

            day: date
                Day to collect data from.

        Returns
        -------
            DataFrame
                Future data.
            str
                Error message.
        """
        first_trade_date = contract.first_trade_date
        last_trade_date = contract.last_trade_date
        if (
            first_trade_date is None
            or day < first_trade_date
            or last_trade_date is None
            or day > last_trade_date
        ):
            message = f"No OHLCV for {contract.ric} on {day.isoformat()}"
            return None, {"message": message}
        dfm, _ = MarketData.__get_future_ohlcv(
            contract.ric, first_trade_date, last_trade_date
        )
        if dfm is not None:
            index = dfm.index == day.isoformat()
            current_day_exists = np.any(index)
            if current_day_exists:
                return dfm.loc[index, :], None
        message = f"No OHLCV for {contract.ric} on {day.isoformat()}"
        return None, {"message": message}

    def get_start_day(self, first_trading_day: date, window: int):
        """
        Get the start day of the strategy.

        Parameters
        ----------
            first_trading_day: date
                First day we have data for.

            window: int
                Number of trading days the strategy need.

        Returns
        -------
            date
                First day the strategy should be started to be traded.
        """
        trading_days = get_calendar("NYSE").valid_days(
            first_trading_day - timedelta(days=2 * window), first_trading_day
        )
        assert len(trading_days) > window
        return trading_days[-window].date()

    def is_trading_day(self, contract: Contract, day: date):
        """
        Check is the day is a trading day for this instrument.

        Parameters
        ----------
            contract: Contract
                Contract object of the instrument.

            day: date
                Day to check.

        Returns
        -------
            bool
                True is the day is a trading day. Folse otherwise.
        """
        try:
            row = self.bardata(contract=contract, day=day)
        except Exception as exception:
            message = str(exception)
            if "No OHLCV for" in message:
                return False
            if "[not-started]" in message:
                return False
            raise exception
        return not np.isnan(row.Close[0])

    def should_roll_today(self, day: date, ticker: str):
        """
        Check is the future needs to be rolled or not.

        Parameters
        ----------
            ticker: str
                Ticker of the instrument.

            day: date
                Day to check.

        Returns
        -------
            bool
                True is the future needs to be rolled.
        """
        front_contract, front_ltd = Contract(day=day, ticker=ticker).front_contract
        if day + timedelta(days=MAXIMUM_NUMBER_OF_DAYS_BEFORE_EXPIRY) < front_ltd:
            return False
        future = FUTURES.get(ticker, {})
        roll_offset_from_reference = timedelta(
            days=future.get("RollOffsetFromReference", -31)
        )

        def is_good_day_to_roll(day, front_contract, next_contract):
            return (
                not is_weekend(day)
                and self.is_trading_day(contract=front_contract, day=day)
                and self.is_trading_day(contract=next_contract, day=day)
            )

        delta = front_ltd - day + roll_offset_from_reference
        next_contract, _ = Contract(day=day, ticker=ticker).front_contract
        for i in range(delta.days, -1, -1):
            _day = day + timedelta(days=i)
            if is_good_day_to_roll(_day, front_contract, next_contract):
                return day == _day
        return False



"""
Market impact access module.
"""
from .contract import Contract
from ..data.constants import FUTURES

DEFAULT_SPREAD = 5e-4


class MarketImpact:
    """
    Market impact.
    """

    def __init__(self):
        pass

    def compute(self):
        """
        Where we will implement the computation of an estimate of the market impact.
        """

    def get(self, ticker=None, ric=None):
        """
        Get the market impact for the instrument.

        Parameters
        ----------
            ticker: str
                Ticker of the instrument.

            ric: str
                RIC of the instrument.

        Returns
        -------
            float
                Market impact for the instrument.
        """
        if ticker is None:
            ticker = Contract(ric=ric).ticker
        return FUTURES.get(ticker, {}).get("Spread", DEFAULT_SPREAD)



"""
Broker simulator
"""
import numpy as np

from .contract import Contract
from .forex import Forex
from .margin import Margin
from .market_data import MarketData
from .market_impact import MarketImpact
from ..data.constants import FUTURES, FUTURE_TYPE


COMMISSION_INTERACTIVE_BROKERS_USD = (
    1.05  # Source: https://www.interactivebrokers.com/en/index.php?f=1590&p=futures2
)


# pylint: disable=too-many-instance-attributes
class Broker:
    """
    Broker. Compute the executions and maintains the positions registry.
    """

    def __init__(
        self,
        cash,
        live,
        no_check=False,
    ):
        self.positions = {
            "Cash": {
                "USD": cash,
            },
            FUTURE_TYPE: {},
        }
        self.previous_close = {}
        self.day = None
        self.executions = []
        self.forex = Forex()
        self.has_execution = False
        self.live = live
        self.margin = Margin()
        self.market_data = MarketData()
        self.market_impact = MarketImpact()
        self.no_check = no_check

    def apply_adjustment(self, adjustment_ratio: float):
        """
        Apply adjustment to position size based on the actual cash on the broker account.

        Parameters
        ----------
            adjustment_ratio: float
                Cash adjustment ratio.

        Returns
        -------
        """
        for key, value in self.positions.items():
            self.positions[key] = {k: v * adjustment_ratio for k, v in value.items()}

    def _apply_commission(self, contract_number: int):
        commission = -np.abs(contract_number) * COMMISSION_INTERACTIVE_BROKERS_USD
        self.positions["Cash"]["USD"] += commission
        return commission

    def _apply_market_impact(self, ric, contract_number, execution_price):
        relative_market_impact = self.market_impact.get(ric=ric)
        ticker = Contract(ric=ric).ticker
        full_point_value = FUTURES[ticker]["FullPointValue"]
        currency = FUTURES[ticker]["Currency"]
        full_point_value_usd = full_point_value * self.forex.to_usd(currency, self.day)
        market_impact = (
            -np.abs(contract_number)
            * relative_market_impact
            * execution_price
            * full_point_value_usd
        )
        self.positions["Cash"][currency] += market_impact
        return market_impact

    def buy_future(self, contract: Contract, contract_number: int):
        """
        Buy future contracts.

        Parameters
        ----------
            contract: Contract
                Instrument Contract.

            contract_number: int
                Number of contracts to buy.

        Returns
        -------
        """
        if not self.live:
            contract_number = np.round(contract_number)
        if contract_number == 0:
            return
        currency = FUTURES[contract.ticker]["Currency"]
        if currency not in self.positions["Cash"]:
            self.positions["Cash"][currency] = 0
        if np.isnan(self.positions["Cash"][currency]):
            raise ValueError("Cash is nan.", ric, self.day)

        row = self.market_data.bardata(contract=contract, day=self.day)
        if np.isnan(row.Close[0]):
            raise ValueError("Close is nan.", contract.ric, self.day)
        execution_price = row.Close[0]
        self.positions[FUTURE_TYPE][contract.ric] = (
            self.positions[FUTURE_TYPE].get(contract.ric, 0) + contract_number
        )
        self.positions["Cash"][currency] -= (
            contract_number
            * execution_price
            * FUTURES[contract.ticker]["FullPointValue"]
        )
        commission = self._apply_commission(contract_number)
        market_impact = self._apply_market_impact(
            contract.ric, contract_number, execution_price
        )
        self._check_initial_margin(contract.ric, contract_number)
        self.executions.append(
            {
                **{
                    "Date": self.day.isoformat(),
                    "Ric": contract.ric,
                    "Ticker": contract.ticker,
                    "Type": "Buy" if contract_number > 0 else "Sell",
                    "ContractNumber": contract_number,
                    "Currency": currency,
                    "ExecutionPrice": execution_price,
                    "FullPointValue": FUTURES[contract.ticker]["FullPointValue"],
                    "Commission": commission,
                    "MarketImpact": market_impact,
                },
                **{f"CashAfter{k}": v for k, v in self.positions["Cash"].items()},
            }
        )
        self.has_execution = True

    def _check_initial_margin(self, ric: str, contract_number: int):
        if self.no_check:
            return
        total_required_margin = 0
        for _ric, _contract_number in self.positions[FUTURE_TYPE].items():
            if _ric == ric:
                _contract_number += contract_number
            ticker = Contract(ric=_ric).ticker
            margin = (
                self.margin.overnight_maintenance_future(ticker, self.day)
                if _ric != ric
                else self.margin.overnight_initial_future(ticker, self.day)
            )
            total_required_margin += np.abs(_contract_number) * margin
        if total_required_margin > self.nav:
            raise Exception(
                f"Initial margin exceeded {self.day.isoformat()} {ric}"
                + f" {contract_number} {total_required_margin} {self.nav}"
            )

    def _check_maintenance_margin(self):
        if self.no_check:
            return
        total_required_margin = 0
        for _ric, _contract_number in self.positions[FUTURE_TYPE].items():
            ticker = Contract(ric=_ric).ticker
            margin = self.margin.overnight_maintenance_future(ticker, self.day)
            if np.isnan(margin):
                continue
            total_required_margin += np.abs(_contract_number) * margin
        if total_required_margin > self.nav:
            raise Exception(f"Maintenance margin exceeded {self.day.isoformat()}")

    def expire_future(self, contract: Contract):
        """
        Force future expiration closing all positions.

        Parameters
        ----------
            ric: str
                Instrument RIC.

        Returns
        -------
        """
        dfm, _ = self.market_data.get_future_ohlcv_for_day(
            contract=contract, day=self.day
        )
        execution_price = (
            dfm.Close[0]
            if not np.isnan(dfm.Close[0])
            else np.nanmedian(dfm[["Open", "High", "Low"]])
        )
        return self.close_future(contract=contract, execution_price=execution_price)

    def close_future(self, contract: Contract, execution_price=None):
        """
        Close future positions.

        Parameters
        ----------
            contract: Contract
                Instrument Contract.

            execution_price: float | None
                Execution price of the position closing.

        Returns
        -------
        """
        currency = FUTURES[contract.ticker]["Currency"]
        if currency not in self.positions["Cash"]:
            self.positions["Cash"][currency] = 0
        if np.isnan(self.positions["Cash"][currency]):
            raise ValueError("Cash is nan.", contract.ric, self.day)
        if execution_price is None:
            row = self.market_data.bardata(contract=contract, day=self.day)
            if np.isnan(row.Close[0]):
                raise ValueError("Close is nan.", contract.ric, self.day)
            execution_price = row.Close[0]
        contract_number = self.positions[FUTURE_TYPE].get(contract.ric, 0)
        self.positions[FUTURE_TYPE][contract.ric] = (
            self.positions[FUTURE_TYPE].get(contract.ric, 0) - contract_number
        )
        self.positions["Cash"][currency] += (
            contract_number
            * execution_price
            * FUTURES[contract.ticker]["FullPointValue"]
        )
        commission = self._apply_commission(contract_number)
        market_impact = self._apply_market_impact(
            contract.ric, contract_number, execution_price
        )
        self.executions.append(
            {
                **{
                    "Date": self.day.isoformat(),
                    "Ric": contract.ric,
                    "Ticker": contract.ticker,
                    "Type": "Close",
                    "ContractNumber": contract_number,
                    "Currency": currency,
                    "ExecutionPrice": execution_price,
                    "FullPointValue": FUTURES[contract.ticker]["FullPointValue"],
                    "Commission": commission,
                    "MarketImpact": market_impact,
                },
                **{f"CashAfter{k}": v for k, v in self.positions["Cash"].items()},
            }
        )
        self.has_execution = True
        return contract_number

    @property
    def nav(self):
        """
        Compute the current Net Asset Value.

        Parameters
        ----------

        Returns
        -------
            float
                NAV
        """
        if np.any([np.isnan(cash) for cash in self.positions["Cash"].values()]):
            raise ValueError("Cash is nan.", self.day)
        cash_in_usd = np.sum(
            [
                value * self.forex.to_usd(currency, self.day)
                for currency, value in self.positions["Cash"].items()
            ]
        )
        nav = cash_in_usd
        for ric, contract_number in self.positions[FUTURE_TYPE].items():
            if contract_number == 0:
                continue
            contract = Contract(ric=ric)
            if self.market_data.is_trading_day(contract=contract, day=self.day):
                row = self.market_data.bardata(contract=contract, day=self.day)
                close = row.Close[0]
                self.previous_close[contract.ric] = close
            else:
                close = self.previous_close.get(ric, np.NaN)
            ticker = Contract(ric=ric).ticker
            full_point_value = FUTURES[ticker]["FullPointValue"]
            currency = FUTURES[ticker]["Currency"]
            full_point_value_usd = full_point_value * self.forex.to_usd(
                currency, self.day
            )
            nav += contract_number * close * full_point_value_usd
        return nav

    def next(self, day):
        """
        Go to next day and set it.

        Parameters
        ----------
            day: date
                Next day.

        Returns
        -------
        """
        self.day = day
        self._check_maintenance_margin()
        self.has_execution = False

    def roll_front_contract(self, ticker):
        """
        For front contract to the next one.

        Parameters
        ----------
            ticker: str
                Instrument ticker.

        Returns
        -------
        """
        front_contract, _ = Contract(day=self.day, ticker=ticker).front_contract
        next_contract, _ = Contract(day=self.day, ticker=ticker).next_contract
        if not self.market_data.is_trading_day(
            contract=front_contract, day=self.day
        ) or not self.market_data.is_trading_day(contract=next_contract, day=self.day):
            return None
        closed_contract_number = 0
        if self.positions[FUTURE_TYPE].get(front_contract.ric, 0) != 0:
            closed_contract_number += self.close_future(contract=front_contract)

        if closed_contract_number != 0:
            self.buy_future(
                contract=next_contract, contract_number=closed_contract_number
            )
        return next_contract.ric

    def sell_future(self, contract: Contract, contract_number: int):
        """
        Sell future contracts.

        Parameters
        ----------
            ric: str
                Instrument RIC.

            contract_number: int
                Number of contracts to buy.

        Returns
        -------
        """
        return self.buy_future(contract, -contract_number)



"""
Definition of the constants of the module
"""
from datetime import date
import json
import os


EMPTY = "empty"

script_path = os.path.abspath(os.path.dirname(__file__))
file_path = os.path.join(script_path, "database-futures.json")
with open(file_path, "r", encoding="utf-8") as handler:
    FUTURES = json.load(handler)

FUTURE_TYPE = "Future"

LAST_MODIFIED = "last-modified"

LETTERS = ["F", "G", "H", "J", "K", "M", "N", "Q", "U", "V", "X", "Z"]

LIBOR_BEFORE_2001 = 6.65125

START_DATE = date(2000, 1, 1)



"""
Forex module.
"""
from datetime import date, datetime

from methodtools import lru_cache
import numpy as np
import pandas as pd

from ..data.client import Client
from ..data.constants import FUTURES


class Forex:
    """
    Forex implementation.
    """

    def __init__(self):
        self._client = Client()

    def bar_to_usd(self, bardata, ticker):
        """
        Convert a bar data to USD values.

        Parameters
        ----------
            bardata: str
                The bardata you convert.

            ticker: str
                The ticker of the instrument. Needed to check its currency.

        Returns
        -------
            DataFrame
                The USD bardata.
        """
        currency = FUTURES[ticker]["Currency"]
        if currency != "USD":
            day = bardata.index[0]
            rate = self.to_usd(currency, day)
            columns = ["Open", "High", "Low", "Close"]
            bardata.loc[:, columns] = bardata.loc[:, columns] * rate
            bardata.loc[:, "Volume"] = bardata.loc[:, "Volume"] / rate
        return bardata

    @lru_cache()
    def __get_forex_ohlcv(self, ric: str, start_date: date, end_date: date):
        """
        Get forex OHLCV.

        Parameters
        ----------
            ric: str
                Instrument RIC.

            start_date: date
                Start date of the time range.

            end_date: date
                End date of the time range.

        Returns
        -------
            DataFrame
                Forex daily OHLCV.
        """
        return self._client.get_daily_ohlcv(ric, start_date, end_date)

    def to_usd(self, currency: str, day: date):
        """
        Get the conversion rate to USD.

        Parameters
        ----------
            currency: str
                Currency code.

            day: date
                Day you want to get forex data for.

        Returns
        -------
            float:
                Conversion rate.
        """
        conversion_rate = np.NaN
        if currency == "AUD":
            conversion_rate = self.__get_pair(day, "USDAUD=R", invert=True)
        elif currency == "CAD":
            conversion_rate = self.__get_pair(day, "CADUSD=R")
        elif currency == "CHF":
            conversion_rate = self.__get_pair(day, "CHFUSD=R")
        elif currency == "EUR":
            conversion_rate = self.__get_pair(day, "USDEUR=R", invert=True)
        elif currency == "GBP":
            conversion_rate = self.__get_pair(day, "USDGBP=R", invert=True)
        elif currency == "HKD":
            conversion_rate = self.__get_pair(day, "HKDUSD=R")
        elif currency == "JPY":
            conversion_rate = self.__get_pair(day, "JPYUSD=R")
        elif currency == "USD":
            conversion_rate = 1
        elif currency == "SGD":
            conversion_rate = self.__get_pair(day, "SGDUSD=R")
        return conversion_rate

    def __get_pair(self, day, ric, invert=False):
        start_date = date(day.year, 1, 1)
        end_date = min(date(day.year, 12, 31), date.today())
        dfm, _ = self.__get_forex_ohlcv(ric, start_date, end_date)
        if dfm is None:
            return np.NaN
        _day = datetime.combine(day, datetime.min.time())
        index = pd.to_datetime(
            dfm.index.map(lambda x: x[0]), format="%Y-%m-%d"
        ).get_indexer([_day], method="nearest")[0]
        dfm = dfm.iloc[index, :]
        return 1 / dfm.Close if invert else dfm.Close



"""
Contract management functions.
"""
from datetime import date, datetime, timedelta

from methodtools import lru_cache

from ..data.client import Client
from ..data.constants import FUTURES, START_DATE


class Contract:
    """
    Contract management.
    """

    def __init__(
        self,
        ric: str = None,
        day: date = None,
        ticker: str = None,
        contract_rank: int = 0,
    ):
        if ric is not None or (day is not None and ticker is not None):
            self.contract_rank = contract_rank
            self.day = day
            self.ric = ric
            self._active_ric = None
            self._client = Client()
            self._first_trade_date = None
            self._last_trade_date = None
            self._ticker = ticker
        else:
            raise Exception("Contract parameters not properly specified.")

    def __str__(self):
        return self.ric

    def __get_ric_year(self):
        if "^" in self.ric:
            year_3 = self.ric.split("^")[1]
            year_4 = self.ric.split("^")[0][-1]
            year_12 = "19" if year_3 in ["8", "9"] else "20"
            year = int(f"{year_12}{year_3}{year_4}")
            return year
        return None

    @lru_cache()
    @staticmethod
    def __get_ric_exists_today(ric, _day=date.today()):
        ric_exists, _ = Client().get_health_ric(ric)
        return ric_exists

    @property
    def active_ric(self):
        """
        Converts to active RIC.

        Parameters
        ----------

        Returns
        -------
            str
                RIC.
        """
        if self._active_ric is None:
            is_recent_ric = self.is_recent_ric
            if is_recent_ric and not self.__get_ric_exists_today(self.ric):
                self._active_ric = self.ric.split("^")[0]
            else:
                self._active_ric = self.ric
        return self._active_ric

    @property
    def is_recent_ric(self):
        """
        Says if a RIC is still active or not.

        Parameters
        ----------

        Returns
        -------
            bool
                True if this is a recent RIC.
        """
        is_recent_ric = False
        year = self.__get_ric_year()
        if year:
            is_recent_ric = year >= (date.today() - timedelta(days=365)).year
        return is_recent_ric

    def __get_contract(self, ticker: str, day: date, contract_rank: int = 0):
        chain = Contract(ticker=ticker, day=day).chain
        contract = chain.iloc[contract_rank, :]
        ltd = datetime.strptime(contract.LTD, "%Y-%m-%d").date()
        ric = contract.RIC
        ric = Contract(ric=ric).active_ric
        return Contract(ric=ric), ltd

    @property
    def front_contract(self):
        """
        Get the front contract characteristics.

        Parameters
        ----------
            ticker: str
                Ticker of the contract.

            day: date
                Date of the contract to be checked.

        Returns
        -------
            date
                Last trading date.

            str
                RIC.
        """
        if self.ticker is None or self.day is None:
            return None, None

        future = FUTURES.get(self.ticker, {})
        roll_offset_from_reference = timedelta(
            days=future.get("RollOffsetFromReference", -31)
        )
        reference_day = self.day - roll_offset_from_reference
        return self.__get_contract(
            ticker=self.ticker, day=reference_day, contract_rank=0
        )

    @property
    def next_contract(self):
        """
        Get the next contract characteristics.

        Parameters
        ----------
            ticker: str
                Ticker of the contract.

            day: date
                Date of the contract to be checked.

        Returns
        -------
            date
                Last trading date.

            str
                RIC.
        """
        future = FUTURES.get(self.ticker, {})
        roll_offset_from_reference = timedelta(
            days=future.get("RollOffsetFromReference", -31)
        )
        reference_day = self.day - roll_offset_from_reference
        return self.__get_contract(
            ticker=self.ticker, day=reference_day, contract_rank=1
        )

    @property
    def ticker(self):
        """
        Convert a RIC to a ticker.

        Parameters
        ----------

        Returns
        -------
            str
                Ticker.
        """
        if self._ticker is None:
            suffix = "^"
            stem_wo_suffix = (
                self.ric.split(suffix)[0] if suffix in self.ric else self.ric
            )
            delayed_data_prefix = "/"
            stem_wo_prefix = (
                stem_wo_suffix.split(delayed_data_prefix)[-1]
                if delayed_data_prefix in stem_wo_suffix
                else stem_wo_suffix
            )
            stem_wo_year = "".join([c for c in stem_wo_prefix if not c.isdigit()])
            stem_wo_month = stem_wo_year[:-1]
            if stem_wo_month in ["SIRT"]:
                return "SI"
            for ticker in FUTURES.keys():
                if stem_wo_month == FUTURES[ticker].get("Stem", {}).get("Reuters"):
                    self._ticker = ticker
        return self._ticker

    @property
    def first_trade_date(self) -> date:
        """
        Get the contract first trade date.

        Parameters
        ----------

        Returns
        -------
            date
                First trade date.
        """
        if self._first_trade_date is None:
            chain = Contract(ticker=self.ticker, day=START_DATE).chain
            if "^" in self.ric:
                contracts = chain.loc[chain.RIC == self.ric, "FTD"]
                if contracts.shape[0] == 0:
                    return None
                self._first_trade_date = datetime.strptime(
                    contracts.iloc[0], "%Y-%m-%d"
                ).date()
            else:
                index = chain.RIC.apply(lambda x: x.split("^")[0]) == self.ric
                contracts = chain.loc[index, :]
                if contracts.shape[0] == 0:
                    return None
                ltd = min(
                    contracts.LTD,
                    key=lambda x: abs(
                        datetime.strptime(x, "%Y-%m-%d").date() - date.today()
                    ),
                )
                ftd = contracts.FTD[contracts.LTD == ltd].iloc[0]
                self._first_trade_date = datetime.strptime(ftd, "%Y-%m-%d").date()
        return self._first_trade_date

    @property
    def last_trade_date(self) -> date:
        """
        Get the contract last trade date.

        Parameters
        ----------

        Returns
        -------
            date
                Last trade date.
        """
        if self._last_trade_date is None:
            chain = Contract(ticker=self.ticker, day=START_DATE).chain
            if "^" in self.ric:
                contracts = chain.loc[chain.RIC == self.ric, "LTD"]
                if contracts.shape[0] == 0:
                    return None
                self._last_trade_date = datetime.strptime(
                    contracts.iloc[0], "%Y-%m-%d"
                ).date()
            else:
                index = chain.RIC.apply(lambda x: x.split("^")[0]) == self.ric
                contracts = chain.loc[index, :]
                if contracts.shape[0] == 0:
                    return None
                ltd = min(
                    contracts.LTD,
                    key=lambda x: abs(
                        datetime.strptime(x, "%Y-%m-%d").date() - date.today()
                    ),
                )
                self._last_trade_date = datetime.strptime(ltd, "%Y-%m-%d").date()
        return self._last_trade_date

    @lru_cache()
    @staticmethod
    def __get_expiry_calendar(ticker: str):
        dfm, err = Client().get_expiry_calendar(ticker)
        return dfm, err

    @property
    def chain(self):
        """
        Get the future contract chain for a given ticker, day and minimum time to expiry.

        Parameters
        ----------

        Returns
        -------
            DataFrame
                Contract chain.
        """
        minimum_time_to_expiry = 0
        dfm, _ = Contract.__get_expiry_calendar(self.ticker)
        if datetime.strptime(
            dfm.LTD.iloc[-1], "%Y-%m-%d"
        ).date() - self.day < timedelta(days=minimum_time_to_expiry):
            expiry_calendar = FUTURES.get(self.ticker, {}).get("ExpiryCalendar", "")
            raise Exception(
                f"Not enough data for {self.ticker}. Download expiry data from {expiry_calendar}"
            )
        index = (dfm.LTD >= self.day.isoformat()) & (
            dfm.WeTrd == 1
        )  # pylint: disable=no-member
        return dfm.loc[index, :].reset_index(drop=True)  # pylint: disable=no-member

    @property
    def will_expire_soon(self):
        """
        Get the future contract chain for a given ticker, day and minimum time to expiry.

        Parameters
        ----------

        Returns
        -------
            bool
                True if the contract expires soon. False otherwise.
        """
        return self.day > self.last_trade_date - timedelta(days=10)


def stem_to_ticker(stem: str):
    """
    Convert a stem to a ticker.

    Parameters
    ----------
        stem: str
            Stem of the contract.

    Returns
    -------
        str
            Ticker.
    """
    tickers = [
        k for k, v in FUTURES.items() if v.get("Stem", {}).get("Reuters") == stem
    ]
    if len(tickers) != 1:
        raise Exception(
            f"No future with Stem.Reuters {stem}. Double check file database-futures.json"
        )
    ticker = tickers[0]
    return ticker
